////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// $Id: GhostScript.cc,v 2.0 1997/10/12 10:21:23 achim Exp $
//                                                                                                                //
// BeDVI                                                                                                          //
// by Achim Blumensath                                                                                            //
// blume@corona.oche.de                                                                                           //
//                                                                                                                //
// This program is free software! It may be distributed according to the GNU Public License (see COPYING).        //
// Some of the code in this file is based on xdvi. The original copyright follows:                                //
//                                                                                                                //
// Copyright (c) 1994 Paul Vojta.  All rights reserved.                                                           //
//                                                                                                                //
// Redistribution and use in source and binary forms, with or without                                             //
// modification, are permitted provided that the following conditions                                             //
// are met:                                                                                                       //
// 1. Redistributions of source code must retain the above copyright                                              //
//    notice, this list of conditions and the following disclaimer.                                               //
// 2. Redistributions in binary form must reproduce the above copyright                                           //
//    notice, this list of conditions and the following disclaimer in the                                         //
//    documentation and/or other materials provided with the distribution.                                        //
//                                                                                                                //
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND                                         //
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE                                          //
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE                                     //
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE                                        //
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL                                     //
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS                                        //
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)                                          //
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT                                     //
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY                                      //
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF                                         //
// SUCH DAMAGE.                                                                                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "DVI.h"

class IOBuffer
{
  private:
    sem_id BufferSem;
    char   *Buffer;
    size_t BufferLen;
    int    InPos;
    int    OutPos;

  public:
    IOBuffer(size_t size = 1024):
      BufferSem(B_ERROR),
      Buffer(NULL),
      BufferLen(size),
      InPos(0),
      OutPos(0)
    {
      BufferSem = create_sem(1, "buffer semaphore");
      Buffer    = new char[size];
    }

    ~IOBuffer()
    {
      if(BufferSem >= B_OK)
      {
        acquire_sem(BufferSem);
        delete_sem(BufferSem);
      }
      if(Buffer)
        delete [] Buffer;
    }

    size_t BytesInBuffer()
    {
      size_t len;

      acquire_sem(BufferSem);

      len = (InPos - OutPos + BufferLen) % BufferLen;

      release_sem(BufferSem);

      return len;
    }

    void Put(char x)
    {
      while(BytesInBuffer() == BufferLen - 1)
        snooze(50000);

      acquire_sem(BufferSem);

      Buffer[InPos] = x;

      InPos = (InPos + BufferLen + 1) % BufferLen;

      release_sem(BufferSem);
    }

    char Get()
    {
      char x;

      while(BytesInBuffer() == 0)
        snooze(50000);

      acquire_sem(BufferSem);

      x = Buffer[OutPos];

      OutPos = (OutPos + BufferLen + 1) % BufferLen;

      release_sem(BufferSem);

      return x;
    }
};

static struct
{
  pid_t    pid;
  int      StdIn[2];       // pipes
  int      StdOut[2];
  IOBuffer InBuffer;
  IOBuffer OutBuffer;

  u_int    PageWidht;      // size of the current page
  u_int    PageHeight;
  int      Magnification;  // magnification currently in use
  int      Shrink;         // shrink factor currently in use

  int      Pending;        // number of ack's we're expecting
  bool     Active;         // if we've started a page yet
  bool     Sending;        // if we're in the middle of send()
  bool     PendingInt;     // if interrupt rec'd while in send()
} GSStatus;


extern PSProcs EmptyProcs;

static bool InitGS(DrawInfo *di);

static	void PSStartup(DrawInfo *di, int xul, int yul, char *cmd)
{
  if(!InitGS(di))
    *di->PS = EmptyProcs;
  else
    di->PS->DrawBegin(di, xul, yul, cmd);
}

static	void NullProc(DrawInfo *di) {}
static	void NullProc2(char *cmd) {}

void DrawBBox(DrawInfo *di);

static void DrawBegin_None(DrawInfo *di, int xul, int yul, char *cmd)
{
  DrawBBox(di);
}

PSProcs EmptyProcs =       // if GhostScript isn't found
{
  NullProc,
  NullProc,
  NullProc,
  NullProc,
  DrawBegin_None,
  NullProc2,
  NullProc2,
  NullProc2
};

PSProcs DVI::PS =          // filled with interface to GhostScript
{
  NullProc,
  NullProc,
  NullProc,
  NullProc,
  PSStartup,
  NullProc2,
  NullProc2,
  NullProc2
};


#if 0

static int32 GSSendThread(void *dummy)
{
  char x;

  while(true)
  {
    x = GSStatus.OutBuffer.Get();

    if(write(GSStatus.StdIn[1], &x, 1) <= 0)
      break;
  }
  return 0;
}

static int32 GSReceiveThread(void *dummy)
{
  char x;

  while(true)
  {
    if(read(GSStatus.StdOut[0], &x, 1) <= 0)
      break;

    GSStatus.InBuffer.Put(x);
  }
  return 0;
}

static	void ReadFromGS()
{
	int	bytes;
	char	*line_end;
	char	*p;

	bytes = read(GS_out, linepos, line + LINELEN - linepos);
	if (bytes < 0) return;
	line_end = linepos + bytes;
	/* Check for ack strings */
	for (p = line; p < line_end - 2; ++p) {
	    p = memchr(p, '\347', line_end - p - 2);
	    if (p == NULL) break;
	    if (memcmp(p, ackstr, 3) == 0) {
		--GS_pending;
		if (debug & DBG_PS)
		    Printf("Got GS ack; %d pending.\n", GS_pending);
		if (p > line) {
		    *p = '\0';
		    Printf("gs: %s\n", line);
		}
		p += 3;
		(void) bcopy(p, line, line_end - p);
		line_end -= p - line;
		linepos = p = line;
		--p;
	    }
	}
	for (;;) {
	    p = memchr(linepos, '\n', line_end - linepos);
	    if (p == NULL) break;
	    *p = '\0';
	    Printf("gs: %s\n", line);
	    ++p;
	    (void) bcopy(p, line, line_end - p);
	    line_end -= p - line;
	    linepos = line;
	}
	linepos = line_end;
	/*
	 * Normally we'd hold text until a newline character, but the buffer
	 * is full.  So we flush it, being careful not to cut up an ack string.
	 */
	if (linepos >= line + LINELEN) {
	    p = line + LINELEN;
	    if ((*--p != '\347' && *--p != '\347' && *--p != '\347')
		    || memcmp(p, ackstr, line + LINELEN - p) != 0)
		p = line + LINELEN;
	    *p = '\0';
	    Printf("gs: %s\n", line);
	    *p = '\347';
	    linepos = line;
	    while (p < line + LINELEN) *linepos++ = *p++;
	}
}

// For handling of SIGPIPE signals from Send()

static	bool sigpipe_error = false;

static	void GSSigPipeHandler(int sig, int code, sigcontext *scp, char *addr)
{
  sigpipe_error = true;
}

static	struct sigaction sigpipe_handler_struct;
  // initialized to {gs_sigpipe_handler, (sigset_t) 0, 0} in InitGS

// This actually sends the bytes to ghostscript.

static	void Send(const char *str, int len)
{
  while(len-- > 0)
    GSStatus.OutBuffer.Put(*str++);
}

/*static void Send(const char *str, int len)
{
  sigaction orig;
  int       bytes;

  if(GSStatus.pid < 0)
    return;

  sigaction(SIGPIPE, &sigpipe_handler_struct, &orig);

  sigpipe_error    = false;
  GSStatus.Sending = true;

#ifndef	STREAMSCONN
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
#endif

  for (;;) {

#ifndef	STREAMSCONN
	    FD_SET(ConnectionNumber(DISP), &readfds);
	    FD_SET(GS_in, &writefds);
	    FD_SET(GS_out, &readfds);

	    if (select(numfds, &readfds, &writefds, (fd_set *) NULL,
		    (struct timeval *) NULL) < 0 && errno != EINTR) {
		perror("select (gs_send)");
		break;
	    }
#else	/* STREAMSCONN *//*
	    for (;;) {
		retval = poll(fds, XtNumber(fds), -1);
		if (retval >= 0 || errno != EAGAIN) break;
	    }
	    if (retval < 0) {
		perror("poll (gs_send)");
		break;
	    }
#endif	/* STREAMSCONN *//*

	    if (XDVI_ISSET(GS_out, &readfds, 1))
		read_from_gs();
	    if (XDVI_ISSET(GS_in, &writefds, 0)) {
		bytes = write(GS_in, cp, len);
		if (bytes == -1) {
		    if (!AGAIN_CONDITION) perror("gs_send");
		}
		else {
		    cp += bytes;
		    len -= bytes;
		    if (len == 0) break;
		}
		if (sigpipe_error) break;
	    }
	    if (XDVI_ISSET(ConnectionNumber(DISP), &readfds, 2)) {
		allow_can = False;
		read_events(False);
		allow_can = True;
		if (GS_pid < 0) break;	/* if timeout occurred *//*
	    }
	}

  // put back generic handler for SIGPIPE
#ifdef	_POSIX_SOURCE
	(void) sigaction(SIGPIPE, &orig, (struct sigaction *) NULL);
#else
	(void) signal(SIGPIPE, orig);
#endif
	GS_sending = False;

	if (sigpipe_error) {
	    Fputs("ghostscript died unexpectedly.\n", stderr);
	    destroy_gs();
	    draw_bbox();
	}

	if (GS_pending_int) {
	    GS_pending_int = False;
	    interrupt_gs();
	}
}*/

/*
 *	Wait for acknowledgement from gs.
 */

static	void WaitAck(int waittime)
{
  struct timeval tv;
  struct timeval tv2;
#ifndef	STREAMSCONN
	struct timeval	*timeout = (struct timeval *) NULL;
#else
	int		timeout	= -1;
	int		retval;
#endif
#if	HAS_SIGIO
	int		oldflags;
#endif

	if (GS_pending == 0) return;
#if	HAS_SIGIO
	oldflags = fcntl(ConnectionNumber(DISP), F_GETFL, 0);
	(void) fcntl(ConnectionNumber(DISP), F_SETFL, oldflags & ~FASYNC);
#endif
	if (waittime != 0) {
	    (void) gettimeofday(&tv, (struct timezone *) NULL);
	    tv.tv_sec += waittime;
#ifndef	STREAMSCONN
	    timeout = &tv2;
#endif
	}
#ifndef	STREAMSCONN
	FD_ZERO(&readfds);
#endif
	while (GS_pending > 0) {
	    if (waittime != 0) {
		(void) gettimeofday(&tv2, (struct timezone *) NULL);
#ifndef	STREAMSCONN
		if (timercmp(&tv2, &tv, >=)) {
		    destroy_gs();
		    break;
		}
		tv2.tv_sec = tv.tv_sec - tv2.tv_sec;
		tv2.tv_usec = tv.tv_usec + 1000000 - tv2.tv_usec;
		if (tv2.tv_usec >= 1000000) tv2.tv_usec -= 1000000;
		else --tv2.tv_sec;
#else
		timeout = 1000 * (int) (tv.tv_sec - tv2.tv_sec)
		    + ((long) tv.tv_usec - (long) tv2.tv_usec) / 1000;
		if (timeout <= 0) {
		    destroy_gs();
		    break;
		}
#endif
	    }
#ifndef	STREAMSCONN
	    FD_SET(ConnectionNumber(DISP), &readfds);
	    FD_SET(GS_out, &readfds);
	    if (select(numfds, &readfds, (fd_set *) NULL, (fd_set *) NULL,
		    timeout) < 0 && errno != EINTR) {
		perror("select (gs_waitack)");
		break;
	    }
#else	/* STREAMSCONN */
	    for (;;) {
		retval = poll(fds + 1, XtNumber(fds) - 1, timeout);
		if (retval >= 0 || errno != EAGAIN) break;
	    }
	    if (retval < 0) {
		perror("poll (gs_waitack)");
		break;
	    }
#endif	/* STREAMSCONN */


	    if (XDVI_ISSET(GS_out, &readfds, 1))
		read_from_gs();
	    if (XDVI_ISSET(ConnectionNumber(DISP), &readfds, 2)) {
		allow_can = False;
		read_events(False);
		allow_can = True;
	    }
	}
#if	HAS_SIGIO
	(void) fcntl(ConnectionNumber(DISP), F_SETFL, oldflags);
#endif
	/* If you bail out here, change the call in interrupt_gs(). */
}

static	void GSToggle(DrawInfo *di)
{
  di->PS->DrawBegin = GSDrawBegin;
}

static	void GSDestroy(DrawInfo *di)
{
  if(LinePos > Line)
  {
    *LinePos = '\0';
    Printf("gs: %s\n", Line);
    LinePos = Line;
  }
  if(GSStatus.pid >= 0)
  {
    kill(GSStatus.pid, SIGKILL);
    GSStatus.pid = -1;
  }
  close(GS_in);
  close(GS_out);
  GSStatus.Active     = false;
  GSStatus.Sending    = false;
  GSStatus.PendingInt = false;
  GSStatus.Pending    = 0;
}

static	void GSInterrupt(DrawInfo *di)
{
  static const char str[] = " stop\n%%xdvimark\n";

  if(GSStatus.Sending)
     GSStatus.PendingInt = true;
  else
  {
    if(GSStatus.Active)
    {
      Send(str, sizeof(str) - 1);
      GSStatus.Active = false;
    }
    di->PS->Interrupt = NullProc;  // prevent deep recursion in WaitAck
    WaitAck(5);
    di->PS->Interrupt = GSInterrupt;
  }
}

static	void GSEndPage(DrawInfo *di)
{
  static const char str[] = "stop\n%%xdvimark\n";

  if(GSStatus.Active)
  {
    Send(str, sizeof(str) - 1);
    GSStatus.Active = false;
    WaitAck(0);
  }
}

static	void GSDrawBegin(DrawInfo *di, int xul, int yul, char *cmd)
{
  static const char str[] = " TeXDict begin\n";

  char Buffer[100];

  // check page_w and page_h to see that they haven't increased

  if(page_w > GSStatus.PageWidth || page_h > GSStatus.PageHeight)
  {
    /* It would be nice if we could just resize the window, but I
     * don't see a convenient way to do that. */
    GSDestroy(di);
  }

  if(GSStatus.pid < 0)
    InitGS(di);

  if(!GSStatus.Active)
  {
    if(magnification != GSStatus.Magnification)
    {
      Sprintf(Buffer, "H TeXDict begin /DVImag %d 1000 div def end stop\n%%%%xdvimark\n", GS_mag = magnification);
      Send(Buffer, strlen(Buffer));
      GS_pending++;
    }
    if(di->DspInfo->ShrinkFactor != GS_shrink)
    {
      Sprintf(Buffer, "H TeXDict begin %d %d div dup /Resolution X /VResolution X end stop\n%%%%xdvimark\n",
              di->DspInfo->PixelsPerInch, GS_shrink = di->DspInfo->ShrinkFactor);
      Send(Buffer, strlen(Buffer));
      GS_pending++;
    }
    Send(str, sizeof(str) - 1);

    GSActive = true;
    GS_pending++;
  }

  Sprintf(Buffer, "%d %d moveto\n", xul, yul);
  Send(Buffer, strlen(Buffer));
  Send(cmd, strlen(cmd));
}

static	void GSDrawRaw(char *cmd)
{
  int len = strlen(cmd);

  if(!GSStatus.Active)
    return;

  cmd[len] = '\n';
  Send(cmd, len + 1);
}

static	void GSDrawFile(char *cmd)
{
  char Buffer[PATH_MAX + 7];

  if(!GSStatus.Active)
    return;

  Sprintf(Buffer, "(%s)run\n", cmd);
  Send(Buffer, strlen(Buffer));
}

static	void GSDrawEnd(char *cmd)
{
  if(!GSStatus.Active)
    return;

  Send(cmd, strlen(cmd));
  Send("\n", 1);
}
#endif

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool InitGS(DrawInfo *di)                                                                                      //
//                                                                                                                //
// initializes the GhostScript interface.                                                                         //
//                                                                                                                //
// DrawInfo *di                         drawing information                                                       //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool InitGS(DrawInfo *di)
{
  // This string reads chunks (delimited by %%xdvimark). The first character of a chunk tells whether a given chunk
  // is to be done within save/restore or not. The `H' at the end tells it that the first group is a header;
  // i.e., no save/restore. `execute' is unique to ghostscript.

  static const char str1[] = "\
/xdvi$run {$error /newerror false put {currentfile cvx execute} stopped pop} def \
/xdvi$ack (\347\310\376) def \
/xdvi$dslen countdictstack def \
{currentfile read pop 72 eq \
    {xdvi$run} \
    {xdvi$run \
      clear countdictstack xdvi$dslen sub {end} repeat } \
  ifelse \
  {(%%xdvimark) currentfile =string {readline} stopped \
    {clear} {pop eq {exit} if} ifelse }loop \
  flushpage xdvi$ack print flush \
}loop\nH";

  static const char str2[] = "matrix currentmatrix \
dup dup 4 get round 4 exch put \
dup dup 5 get round 5 exch put setmatrix\n\
stop\n%%xdvimark\n";

  static const char *GSargv[] =
  {
    "gs",
    NULL,                  // filled with "-r%d"
    "-sDEVICE=bitrgb",
    "-sGrayValues=256",
    "-dNOPAUSE",
    "-dSAFER",
    "-q",
    "-sOutputFile=-",
    "-",
    NULL
  };

  char Buffer[100];

#if 0
  gs_atom = XInternAtom(DISP, "GHOSTVIEW", False);
	/* send bpixmap, orientation, bbox (in pixels), and h & v resolution */
	Sprintf(buf, "%ld %d 0 0 %u %u 72 72",
	    None,			/* bpixmap */ 
	    XtPageOrientationPortrait,
	    GS_page_w = page_w, GS_page_h = page_h);
	XChangeProperty(DISP, mane.win, gs_atom, XA_STRING, 8,
	    PropModeReplace, (unsigned char *) buf, strlen(buf));

	gs_colors_atom = XInternAtom(DISP, "GHOSTVIEW_COLORS", False);
	Sprintf(buf, "%s %ld %ld", "Color", fore_Pixel, back_Pixel);
	XChangeProperty(DISP, mane.win, gs_colors_atom, XA_STRING, 8,
	    PropModeReplace, (unsigned char *) buf, strlen(buf));

	XSync(DISP, False);		/* update the window */

  if(pipe(GSStatus.StdIn) != 0 || pipe(GSStatus.StdOut) != 0)
    return false;

  GSStatus.pid = fork();

  if(GSStatus.pid == 0)  // child
  {
    close(GSStatus.StdIn[1]);
    dup2(GSStatus.StdIn[0], 0);
    close(GSStatus.StdIn[0]);
    close(GSStatus.StdOut[0]);

    dup2(GSStatus.StdOut[1], 1);
    dup2(GSStatus.StdOut[1], 2);
    close(GSStatus.StdOut[1]);

    sprintf(Buffer, "-r%d", di->DspInfo->PixelsPerInch / di->DspInfo->ShrinkFactor);
    GSargv[1] = Buffer;

    execvp(GSargv[0], GSargv);
    exit(1);
  }
  if(GS_pid == -1)  // error
    return false;

  close(GSStatus.StdIn[0]);
  close(GSStatus.StdOut[1]);

  // Set std_in for non-blocking I/O

  fcntl(GSStatus.StdIn[1], F_SETFL, fcntl(GSStatus.StdIn[1], F_GETFL, 0) | O_NONBLOCK);

  sigpipe_handler_struct.sa_handler = gs_sigpipe_handler;
  sigemptyset(&sigpipe_handler_struct.sa_mask);

#ifndef	STREAMSCONN
	numfds = ConnectionNumber(DISP);
	if (numfds < std_in[1]) numfds = std_in[1];
	if (numfds < std_out[0]) numfds = std_out[0];
	++numfds;
#else	/* STREAMSCONN */
	fds[0].fd = std_in[1];
	fds[1].fd = std_out[0];
	fds[2].fd = ConnectionNumber(DISP);
#endif	/* STREAMSCONN */

  *di->PS = GSProcs;

  GSStatus.Active        = false;
  GSStatus.Sending       = false;
  GSStatus.PendingInt    = false;
  GSStatus.Pending       = 1;
  GSStatus.Magnification = -1;
  GSStatus.Shrink        = -1;

  Send(str1, sizeof(str1) - 1);
  Send(psheader, psheaderlen);
  Sprintf(Buffer, "[1 0 0 -1 0 %d] concat\n", page_h);
  Send(Buffer, strlen(Buffer));
  Send(str2, sizeof(str2) - 1);
  WaitAck(0);

  if(GSStatus.pid < 0)   // if something happened 
  {
    GSDestroy(di);
    return false;
  }

  if(!postscript)
    GSToggle(di);	/* if we got a 'v' already */
  else
  {
    CanIt = true;		/* ||| redraw the page */
    longjmp(CanItEnv, 1);
  }
  return true;
#endif
  return false;
}
