////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// $Id: TeXFont.cc,v 2.0 1997/10/12 10:21:39 achim Exp $
//                                                                                                                //
// BeDVI                                                                                                          //
// by Achim Blumensath                                                                                            //
// blume@corona.oche.de                                                                                           //
//                                                                                                                //
// This program is free software! It may be distributed according to the GNU Public License (see COPYING).        //
// Some of the code in this file is based on xdvi. The original copyright follows:                                //
//                                                                                                                //
// Copyright (c) 1994 Paul Vojta.  All rights reserved.                                                           //
//                                                                                                                //
// Redistribution and use in source and binary forms, with or without                                             //
// modification, are permitted provided that the following conditions                                             //
// are met:                                                                                                       //
// 1. Redistributions of source code must retain the above copyright                                              //
//    notice, this list of conditions and the following disclaimer.                                               //
// 2. Redistributions in binary form must reproduce the above copyright                                           //
//    notice, this list of conditions and the following disclaimer in the                                         //
//    documentation and/or other materials provided with the distribution.                                        //
//                                                                                                                //
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND                                         //
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE                                          //
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE                                     //
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE                                        //
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL                                     //
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS                                        //
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)                                          //
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT                                     //
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY                                      //
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF                                         //
// SUCH DAMAGE.                                                                                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <InterfaceKit.h>
#include "defines.h"
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <Debug.h>

extern "C"
{
  #define string _string
  #include "kpathsea/tex-glyph.h"
  #undef string
}

#include "BeDVI.h"
#include "DVI-View.h"
#include "TeXFont.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// Font::Font(DVI *doc, const char *name, float size, long chksum, int magstep, double dimconvert)                //
//                                                                                                                //
// Initializes a Font.                                                                                            //
//                                                                                                                //
// DVI        *doc                      document the font appears in                                              //
// const char *name                     name of the font                                                          //
// float      size                      size of the font                                                          //
// long       chksum                    checksum                                                                  //
// int        magstep                   magnification                                                             //
// double     dimconvert                factor to convert dimensions                                              //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Font::Font(DVI *doc, const char *name, float size, long chksum, int magstep, double dimconvert):
  Name(NULL),
  Size(size),
  ChkSum(chksum),
  MagStep(magstep),
  DimConvert(dimconvert),
  UseCount(0),
  Loaded(false),
  Virtual(false),
  VFTable()
{
  if(name)
    if(Name = new char[strlen(name) + 1])
      strcpy(Name, name);

  Load(doc);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// Font::~Font()                                                                                                  //
//                                                                                                                //
// Deletes a font.                                                                                                //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Font::~Font()
{
  if(Name)
    delete [] Name;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool Font::Open(char *&FontFound, int &SizeFound)                                                              //
//                                                                                                                //
// Opens the file of a font.                                                                                      //
//                                                                                                                //
// char *&FontFound                     used to return the name of the font found                                 //
// int  &SizeFound                      used to return the size of the font found                                 //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Font::Open(char *&FontFound, int &SizeFound)
{
  char *name;

  acquire_sem(kpse_sem);
  name = kpse_find_vf(Name);

  if(name)
  {
    SizeFound = Size;
    FontFound = NULL;
  }
  else
  {
    kpse_glyph_file_type FileResult;

    name = kpse_find_glyph(Name, (u_int)(Size + 0.5), kpse_any_glyph_format, &FileResult);

    if(name)
    {
      if(FileResult.source == kpse_glyph_source_fallback)
        FontFound = FileResult.name;
      else
        FontFound = NULL;

      SizeFound = FileResult.dpi;
    }
  }
  release_sem(kpse_sem);

  if(!name)
    return false;

  if(File.SetTo(name, O_RDONLY) != B_NO_ERROR)
    return false;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void Font::ReallocFont(u_char num)                                                                             //
//                                                                                                                //
// shrinks the arrays containing font data to their minimum size.                                                 //
//                                                                                                                //
// u_char num                           used length of array                                                      //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Font::ReallocFont(u_char num)
{
  try
  {
    if(Virtual)
    {
      Macro *m;

      if(!(m = new Macro[num]))
        return;

      memcpy(m, Macros, num * sizeof(Macro));

      bzero(Macros, 256 * sizeof(Macro));
      delete [] Macros;

      Macros = m;
    }
    else
    {
      Glyph *g;

      if(!(g = new Glyph[num]))
        return;

      memcpy(g, Glyphs, num * sizeof(Glyph));

      bzero(Glyphs, 256 * sizeof(Glyphs));
      delete [] Glyphs;

      Glyphs = g;
    }
  }
  catch(...)
  {
    TRACE();
    return;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool Font::Load(DVI *doc)                                                                                      //
//                                                                                                                //
// Loads a font.                                                                                                  //
//                                                                                                                //
// DVI        *doc                      document the font appears in                                              //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Font::Load(DVI *doc)
{
  char  *FontFound;
  int   SizeFound;
  u_int Type;

  try
  {
    if(!Open(FontFound, SizeFound))
      return false;

    if(FontFound)
    {
      if(Name)
        delete [] Name;
      Name = FontFound;
    }

    Size    = SizeFound;
    MaxChar = 255;
    SetChar = DVI::SetChar;

    Type = ReadInt(&File, 2);

    if(Type == PK_Magic)
    {
      if(!ReadPKIndex(doc, this))
        return false;
    }
    else if(Type == GF_Magic)
    {
      if(!ReadGFIndex(doc, this))
        return false;
    }
    else if(Type == VF_Magic)
    {
      if(!ReadVFIndex(doc, this))
        return false;

      SetChar = DVI::SetVFChar;
    }
    else
      return false;

    if(Virtual)
      while(MaxChar > 0 && Macros[MaxChar].Position == NULL)
        MaxChar--;
    else
      while(MaxChar > 0 && Glyphs[MaxChar].Addr == 0)
        MaxChar--;

    if(MaxChar < 255)
      ReallocFont(MaxChar + 1);

    Loaded = true;

    return true;
  }
  catch(...)
  {
    TRACE();
    return false;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void Font::FlushShrinkedGlyphes()                                                                              //
//                                                                                                                //
// Frees all bitmaps of shrinked glyphes which is neccessary if the anti aliasing method is changed.              //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Font::FlushShrinkedGlyphes()
{
  int i;

  for(i = 0; i < MaxChar; i++)
    if(Glyphs[i].SBitMap)
    {
      delete Glyphs[i].SBitMap;
      Glyphs[i].SBitMap = NULL;
    }
}


/* Glyph **********************************************************************************************************/


u_char *Glyph::ColourTable[MaxShrinkFactor + 1] = {};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// Glyph::Glyph()                                                                                                 //
//                                                                                                                //
// Initializes a Glyph.                                                                                           //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Glyph::Glyph():
  Addr(0),
  Advance(0),
  Ux(0), Uy(0),
  UBitMap(NULL),
  Sx(0), Sy(0),
  SBitMap(NULL)
{
  static int32 TableInitialized = 0;           // record, if `ColourTable' is already initialized

  if(atomic_add(&TableInitialized, 1) < 1)
  {
    // initialize table

    BScreen scr;
    int     Colour;

    for(int Factor = 2; Factor < MaxShrinkFactor; Factor++)
    {
      ColourTable[Factor] = new u_char [Factor * Factor + 1];

      for(int i = 0; i <= Factor * Factor; i++)
      {
        Colour = 255 - (510 * i + Factor * Factor) / (2 * Factor * Factor);
        ColourTable[Factor][i] = scr.IndexForColor(Colour, Colour, Colour);
      }
    }
  }
  else
    atomic_add(&TableInitialized, -1);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// Glyph::~Glyph()                                                                                                //
//                                                                                                                //
// Deletes a Glyph.                                                                                               //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Glyph::~Glyph()
{
  delete UBitMap;
  delete SBitMap;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool Glyph::Shrink(int Factor)                                                                                 //
//                                                                                                                //
// shrinks a glyph.                                                                                               //
//                                                                                                                //
// int Factor                           factor the glyph should be shrinked                                       //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Glyph::Shrink(int Factor, bool AntiAliasing)
{
  if(AntiAliasing)
    return ShrinkGrey(Factor);
  else
    return ShrinkMonochrome(Factor);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool Glyph::ShrinkMonochrome(int Factor)                                                                       //
//                                                                                                                //
// shrinks a glyph without antialiasing.                                                                          //
//                                                                                                                //
// int Factor                           factor the glyph should be shrinked                                       //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Glyph::ShrinkMonochrome(int Factor)
{
  BRect      r;
  int        Width;
  int        Height;
  int        ShrunkBytesWide;
  int        UnshrunkBytesWide;
  int        RowsLeft;
  int        Rows;
  int        ColsLeft;
  int        Cols;
  int        InitCols;
  BitmapUnit *OldPtr;
  BitmapUnit *NewPtr;
  BitmapUnit m;
  BitmapUnit *cp;
  int        MinSample = 0.4 * Factor * Factor;

  if(SBitMap)         // already shrunken?
    return true;

  try
  {
    Sx       = Ux / Factor;
    InitCols = Ux - Sx * Factor;

    if(InitCols <= 0)
      InitCols += Factor;
    else
      Sx++;

    Cols = Uy + 1;
    Sy   = Cols / Factor;
    Rows = Cols - Sy * Factor;

    if(Rows <= 0)
    {
      Rows += Factor;
      Sy--;
    }

    r  = UBitMap->Bounds();

    Width  = r.IntegerWidth()  + 1;
    Height = r.IntegerHeight() + 1;

    r.left   = 0;
    r.top    = 0;
    r.right  = ((Sx + (Width  - Ux   + Factor - 1) / Factor + BITS_PER_UNIT - 1) & ~(BITS_PER_UNIT - 1)) - 1;
    r.bottom = Sy + (Height - Cols + Factor - 1) / Factor;

    if(!(SBitMap = new BBitmap(r, B_MONOCHROME_1_BIT)))
      return false;

    OldPtr = (BitmapUnit *)UBitMap->Bits();
    NewPtr = (BitmapUnit *)SBitMap->Bits();

    ShrunkBytesWide   = SBitMap->BytesPerRow();
    UnshrunkBytesWide = UBitMap->BytesPerRow();
    RowsLeft          = Height;

    bzero(NewPtr, SBitMap->BitsLength());

    while(RowsLeft)
    {
      if(Rows > RowsLeft)
        Rows = RowsLeft;

      ColsLeft = Width;
      cp       = NewPtr;
      Cols     = InitCols;

#ifdef MSB_FIRST
      m = (BitmapUnit) 1 << (BITS_PER_UNIT - 1);
#else
      m = 1;
#endif

      while(ColsLeft)
      {
        if(Cols > ColsLeft)
          Cols = ColsLeft;

        if(Sample(OldPtr, UnshrunkBytesWide, Width - ColsLeft, Cols, Rows) >= MinSample)
          *cp |= m;

#ifdef MSB_FIRST
        if(m == 1)
        {
          m = (BitmapUnit) 1 << (BITS_PER_UNIT - 1);
          cp++;
        }
        else
          m >>= 1;
#else
        if(m == (BitmapUnit) 1 << (BITS_PER_UNIT - 1))
        {
          m = 1;
          cp++;
        }
        else
          m <<= 1;
#endif

        ColsLeft -= Cols;
        Cols      = Factor;
      }
      NewPtr   =  BMU_ADD(NewPtr, ShrunkBytesWide);
      OldPtr   =  BMU_ADD(OldPtr, Rows * UnshrunkBytesWide);
      RowsLeft -= Rows;
      Rows     =  Factor;
    }

    Sy = Uy / Factor;

    return true;
  }
  catch(...)
  {
    TRACE();

    delete SBitMap;
    SBitMap = NULL;

    return false;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool Glyph::ShrinkGrey(int Factor)                                                                             //
//                                                                                                                //
// shrinks a glyph with antialiasing.                                                                             //
//                                                                                                                //
// int Factor                           factor the glyph should be shrinked                                       //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Glyph::ShrinkGrey(int Factor)
{
  BRect      r;
  int        Width;
  int        Height;
  int        ShrunkBytesWide;
  int        UnshrunkBytesWide;
  int        RowsLeft;
  int        Rows;
  int        ColsLeft;
  int        Cols;
  int        InitCols;
  BitmapUnit *OldPtr;
  uint8      *NewPtr;
  uint8      *cp;

  if(SBitMap)         // already shrunken?
    return true;

  try
  {
    Sx       = Ux / Factor;
    InitCols = Ux - Sx * Factor;

    if(InitCols <= 0)
      InitCols += Factor;
    else
      Sx++;

    Cols = Uy + 1;
    Sy   = Cols / Factor;
    Rows = Cols - Sy * Factor;

    if(Rows <= 0)
    {
      Rows += Factor;
      Sy--;
    }

    r  = UBitMap->Bounds();

    Width  = r.IntegerWidth()  + 1;
    Height = r.IntegerHeight() + 1;

    r.left   = 0;
    r.top    = 0;
    r.right  = Sx + (Width  - Ux   + Factor - 1) / Factor - 1;
    r.bottom = Sy + (Height - Cols + Factor - 1) / Factor;

    if(!(SBitMap = new BBitmap(r, B_COLOR_8_BIT)))      // should be replaced by B_GRAYSCALE_8_BIT
      return false;

    OldPtr = (BitmapUnit *)UBitMap->Bits();
    NewPtr = (uint8 *)SBitMap->Bits();

    ShrunkBytesWide   = SBitMap->BytesPerRow();
    UnshrunkBytesWide = UBitMap->BytesPerRow();
    RowsLeft          = Height;

    memset(NewPtr, 255, SBitMap->BitsLength());

    while(RowsLeft)
    {
      if(Rows > RowsLeft)
        Rows = RowsLeft;

      ColsLeft = Width;
      Cols     = InitCols;
      cp       = NewPtr;

      while(ColsLeft)
      {
        if(Cols > ColsLeft)
          Cols = ColsLeft;

        *cp++ = ColourTable[Factor][Sample(OldPtr, UnshrunkBytesWide, Width - ColsLeft, Cols, Rows)];

        ColsLeft -= Cols;
        Cols      = Factor;
      }
      NewPtr   += ShrunkBytesWide;
      OldPtr   =  BMU_ADD(OldPtr, Rows * UnshrunkBytesWide);
      RowsLeft -= Rows;
      Rows     =  Factor;
    }

    Sy = Uy / Factor;

    return true;
  }
  catch(...)
  {
    TRACE();

    delete SBitMap;
    SBitMap = NULL;

    return false;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// int Glyph::Sample(BitmapUnit *Bits, int BytesPerRow, int BitSkip, int Width, int Height)                       //
//                                                                                                                //
// Counts the number of black pixels in a rectangle which should be shrinked to one pixel.                        //
//                                                                                                                //
// BitmapUnit *Bits                     pointer to the top row of the rectangle in the bitmap                     //
// int        BytesPerRow               bytes per row of the bitmap                                               //
// int        BitSkip                   left edge of the rectangle                                                //
// int        Width                     width of the rectangle                                                    //
// int        Height                    height of the rectangle                                                   //
//                                                                                                                //
// Result:                              number indicating color of the pixel                                      //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Glyph::Sample(BitmapUnit *Bits, int BytesPerRow, int BitSkip, int Width, int Height)
{
  static const int SampleCount[] =
  {
    0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6
  };

  BitmapUnit *ptr;
  BitmapUnit *EndPtr;
  BitmapUnit *cp;
  int        BitsLeft;
  int        n;
  int        BitShift;
  int        wid;

  ptr      = Bits + BitSkip / BITS_PER_UNIT;
  EndPtr   = BMU_ADD(Bits, Height * BytesPerRow);
  BitsLeft = Width;
  n        = 0;

#ifdef MSB_FIRST
  BitShift = BITS_PER_UNIT - BitSkip % BITS_PER_UNIT;
#else
  BitShift = BitSkip % BITS_PER_UNIT;
#endif

  while(BitsLeft)
  {
#ifdef MSB_FIRST
    wid = BitShift;
#else
    wid = BITS_PER_UNIT - BitShift;
#endif

    if(wid > BitsLeft)
      wid = BitsLeft;

    if(wid > 6)
      wid = 6;

#ifdef MSB_FIRST
    BitShift -= wid;
#endif

    for(cp = ptr; cp < EndPtr; cp = BMU_ADD(cp, BytesPerRow))
      n += SampleCount[(*cp >> BitShift) & BitMasks[wid]];

#ifdef MSB_FIRST
    if(BitShift == 0)
    {
      BitShift = BITS_PER_UNIT;
      ptr++;
    }
#else
    BitShift += wid;
    if(BitShift == BITS_PER_UNIT)
    {
      BitShift = 0;
      ptr++;
    }
#endif

    BitsLeft -= wid;
  }
  return n;
}
